/*  Program name: "F"
    Compiled under Microsoft C 6.00 and DOS 4.01, May 9, 1990
    Thu May 10, 1990, John Jarvis
    4111 Ave. A., Apt. 209, Austin, TX 78751  512-452-4755
    Released "as is" to public domain without charge or liability.

    Purpose: DIR, DELETE and RMDIR utility program, with enhancements.

             Enhancements:
                1. Redundant "*" wildcard is unnecessary -- input commands
                   are assumed to be wild cards.  Up to two tokens allowed.
                2. Option to OMIT files by wildcard is allowed
                3. Displays HIDDEN, SYSTEM and VOLID files
                4. Displays file size but supresses file dates to reduce clutter
                5. Separates directory names from other file names,
                   and displays both groups in alphabetical order
                6. Does NOT report or count the ever-present "." and ".."
                   directories which should be invisible
                7. Directory names are highlighted in reverse-video which
                   is not ANSI dependent (see notes below)
                8. Full pathnames allowed
                9. Permission is asked after selected files are displayed,
                   but before deletes begin.  Default is NO.  This utility
                   let's you change your mind before it's too late!
               10. Full directories, hidden files, system files or volid
                   files will NOT be deleted, even if requested
               11. Empty directories and read only files can be deleted
                   if desired
               12. Automatic paging (24 lines per page plus "more..." prompt)
                   Paging can be disabled for redirected batch listings.
               14. Long listings can be cancelled using Ctrl-C

             Limitations:
                1000 files max (warning will be given but 1000 will be done)
                Does not recursively list or delete contents of subdirectories
                Deletes will not be done when display is redirected to disk, be-
                    cause permission to delete must be given manually.  This is
                    not necessarily bad.  Program will hang up though, while
                    waiting for "y" key or any other key to be pressed.
                Does not append file notes like Peter Norton FI command

    Usage: f [/omit_template] [/k/n] [file_template] [file_template2]
             (relative position of switch in command string is not critical)
             (either "/" or "-" are acceptable switches)

    Examples: f        -- lists all files in working directory, except for
                          the annoying "." and ".." files.
              f \      -- list all files in root directory, including the
                             hidden, system and volid files
              f .d     -- would show a.dat, b.dmp, ...
              f ile    -- would show <DIR> SMILE, while.c, mile.bin, f.ile, ...
              f a .ext -- would show fab.ext, a.ext, aaa.ext, ...
              f /.doc  -- show all files except any containing ".doc"
              f .doc /big.doc -- shows any "*.doc" EXCEPT "big.doc"
              f \a\.c  -- list all files matching \a\*.c
              f /k /precious .c -- list and then Kill, with permission asked,
                                   all *.c files, except precious.c
              f > disklog -n      -- List all files to disk, without paging

    Subdirectories which match wildcards are listed first, if any, then
    any non-directory files.

    Paging (24 lines per page) is automatic -- press the spacebar to
    continue for long listings.  When long outputs are redirected to disk
    by using "f > disk.ext ...", use the /n switch to disable the pause for
    each page.

    This is the complete source listing, except for the BIOS calls in the
    CXL library routines called by RSAY and ERROR, to write reverse video
    message to the screen, independent of any ANSI driver (as suggested by
    Wally Sheldon of The Programmers' Retreat BBS, Austin, TX).  To build
    this program without the CXL libraries (available as shareware from
    Mike Smedley in San Antonio, BBS phone 512-590-0460), just edit RSAY
    and ERROR to use ANSI calls (reverse video on = ESC[7m, reverse
    video off = ESC[m), or just take out the reverse video calls competely.

MSC build command is:
    cl /F 2100 f cxlmss.lib         (stacksize ox2100 is for 1000 filmax)
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <malloc.h>
#include <direct.h>

#include <core.h>
#include <cxlvid.h>   /* for cxl readcur and gotoxy_ functions */

void burnup();
void rsay(char *fmt, ...);
void error(char *fmt, ...);
int flist(char *template, char **filptr, unsigned char *atribz, int filmax);
int getswi(int *argc, char **argv, char swiz[]);
void psort(int *pary, char **linptr, int nlins);
void dirx(char *template, char *dirpath);
int getyn(int yes1);

#define FILMAX 1000
int row, col, burned;

main(int argc, char **argv)
{
    int j, k, nfil, nfnd, cnt, lenext, row, col;
    long flen;
    char tag1[13], tag2[13], fil[13];
    char cwd[40];
    char ext[4];
    char swiz[20];
    char *filptr[FILMAX];
    char *hitptr[FILMAX];
    int pary[FILMAX];
    unsigned char atrib, atribz[FILMAX];
    char dirpath[60];
    char template[60];
    char *p;
    FILE *f1;
    int ntag = 0;
    int omit = 0;
    int kflag = 0;
    int special = 0;
    int page = 1;

    burned = 0;
    if ((omit = getswi(&argc, argv, swiz))) {
        if ((p = strstr(swiz,"/k/"))) {         /* extract /k flag */
            kflag = 1;
            omit--;
            strcpy(p,p+2);
        } else if ((p = strstr(swiz,"/k")) == swiz + strlen(swiz) -2) {
            kflag = 1;
            omit--;
            *p = '\0';
        }
        if ((p = strstr(swiz,"/n/"))) {         /* extract no page flag */
            page = 0;
            omit--;
            strcpy(p,p+2);
        } else if ((p = strstr(swiz,"/n")) == swiz + strlen(swiz) -2) {
            page = 0;
            omit--;
            *p = '\0';
        }
    }

    if (argc >= 2) {
        ntag = 1;
        strcpy(tag1, strlwr(argv[1]));
        if (strchr(tag1,'*'))
            error("Error: \"*\" is not a legal filename character\n");
    }
    if (argc > 2) {
        ntag = 2;
        strcpy(tag2, strlwr(argv[2]));
        if (strchr(tag2,'*'))
            error("Error: \"*\" is not a legal filename character\n");
    }

    dirx(tag1,dirpath);
    strcpy(template,dirpath);
    if (*(template + strlen(template) - 1) != '\\')
        strcat(template,"\\");
    strcat(template,"*.*");
    nfil = flist(template, filptr, atribz, FILMAX);
    psort( pary, filptr, nfil);

    cnt = 0;
    nfnd = 0;

    /* strip out dirpath from tag1, if present */
    if ((p = strrchr(tag1,'\\')))
        strcpy(tag1,p+1);

/* first pass -- report any directories */
    for (j=0; j<nfil; j++) {
        strcpy(fil, strlwr(filptr[j]));
        if (omit && strstr(fil,swiz+1))
            *fil = '\0';
        if (strlen(fil) && (ntag==0 || strstr(fil,tag1)))
            if (ntag <= 1 || strstr(fil,tag2)) {
                atrib = atribz[pary[j]];
                if (atrib & 16) {
                    strupr(fil);
                    rsay("<DIR>");
                    printf(" %s\n", fil);
                    nfnd++;
                }
            }
    }
    printf("\n");

/* second pass -- non-directory files */
    for (j=0; j<nfil; j++) {
        strcpy(fil, strlwr(filptr[j]));
        if (omit && strstr(fil,swiz+1))
            *fil = '\0';
        if (strlen(fil) && (ntag==0 || strstr(fil,tag1)))
            if (ntag <= 1 || strstr(fil,tag2)) {
                atrib = atribz[pary[j]];
                if (atrib != 16) {
                    if (atrib & 14)
                        special++;  /* 2,4,8 = hidden, system, volid */
                    strcpy(template,dirpath);
                    if (*(template + strlen(template) - 1) != '\\')
                        strcat(template,"\\");
                    strcat(template,fil);
                    f1 = fopen(template,"r");
                    flen = filelength(fileno(f1));
                    fclose(f1);
                    if ((p = strchr(fil,'.'))) {
                        strcpy(ext,p+1);
                        *p = '\0';
                    } else
                        *ext = '\0';
                    if (strlen(fil) < 8)
                        printf("%s\t ", fil);
                    else
                        printf("%s ", fil);
                    lenext = strlen(ext);
                    if (lenext == 0)
                        printf("   ");
                    else if (lenext == 1)
                        printf("%s  ", ext);
                    else if (lenext == 2)
                        printf("%s ", ext);
                    else
                        printf("%s", ext);

                    if (flen > 999999)
                        printf(" %2ld,%3.3ld,%3.3ld ", flen / 1000000,
                           (flen % 1000000) / 1000, flen % 1000);
                    else if (flen > 999)
                        printf("    %3ld,%3.3ld ", flen / 1000, flen % 1000);
                    else
                        printf("        %3ld ", flen);

                    if ((atrib = atribz[pary[j]]) & 1)
                        printf(" ro");
                    if (atrib & 2)
                        printf(" hidden");
                    if (atrib & 4)
                        printf(" system");
                    if (atrib & 8)
                        printf(" volid");
                    if (atrib & 32)
                        printf(" a");
                    printf("\n");

                    nfnd++;
                    if (page && j < nfil-1 && nfnd % 23 == 0) {
                        printf("                         more...");
                        getch();
                        printf("\n");
                    }
                }
            }
    }
    printf("\n");
    if (ntag == 0)
        printf("%d files in %s\n", nfil, dirpath);
    else if (ntag == 1)
        printf("%d of %d files match \"%s\" in %s\n", nfnd, nfil, tag1,
            dirpath);
    else
        printf("%d of %d files match \"%s\"\ and \"%s\" in %s\n",
            nfnd, nfil, tag1, tag2, dirpath);
    if (nfnd == FILMAX)
        printf("Program limit of %d files was reached -- there may be more\n",
            FILMAX);

    if (!kflag || !nfnd)
        exit(0);

    if (special)
        printf("Delete ALL of the above, except for %d special files ?",
            special);
    else
        printf("Delete ALL of the above ?");
    if (!getyn(0))
        exit(0);

    for (j=0; j<nfil; j++) {
        strcpy(fil, strlwr(filptr[j]));
        if (omit && strstr(fil,swiz+1))
            *fil = '\0';
        if (strlen(fil) && (ntag==0 || strstr(fil,tag1)))
            if (ntag <= 1 || strstr(fil,tag2)) {
                atrib = atribz[pary[j]];
                if (atrib & 14)
                    continue;
                if (atrib & 1)
                    _dos_setfileattr(fil,_A_NORMAL);
                if (atrib == 16) {
                    if (rmdir(fil))
                        rsay("Directory %s not deleted -- not empty\n", fil);
                    else
                        burnup();
                } else {
                    if (remove(fil))
                        rsay("Could not delete %s\n", fil);
                    else
                        burnup();
                }
            }
    }
    printf("\n");
}


void burnup()
{
    burned++;
    if (burned == 1) {
        printf("Deleted    1");
        readcur(&row,&col);
    } else {
        gotoxy_(row,8);
        printf("%4d", burned);
    }
}


/*  rsay.c

    Usage:  rsay("too many %d widgets\n", widsum);
            rsay("%s %d %s\n", "still too many", widsum, "widgets");
            rsay("%s", buf);

    This uses the CXL library revattr function to get reverse video which
    is independent of ANSI driver (revattr uses BIOS calls).

    See also:
        warning -- same as rsay, except does an EXIT(1)  \n assumed
        error   -- same as rsay, except does an EXIT(2)  \n assumed

    All three -- rsay, warning and error -- accept variable argument lists.

    However, "\n" is explicit in rsay, but implicit in warning or error.
    If a string is terminated by "\n" in rsay, it is NOT printed in reverse
    video, but a newline is done after the reverse tag is displayed.

    In order to detect if a last newline character was passed, the
    variable argument list is first written to an 80 char buffer, then
    the buffer is written out, and then the attributes of the written
    characters are reversed using BIOS calls.

    This is a little more complicated than WARNING or ERROR.
    If a "\n" character is internal to the string, what happens is that
    only the last part of the string gets highlighted -- because the jump
    back to do the reverse video does not jump back to the former row.
    If a "\n" character is the first character, it has the effect of a
    normal line feed before the message begins.  The revsere video strip
    is always limited to only one line.
*/

#include <stdio.h>
#include <stdarg.h>
#include <core.h>
#include <string.h>

void rsay(char *fmt, ...)
{
    int len, row, col, row2, col2;
    int nlflag = 0;
    char buf[80];

    va_list args;
    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);

    len = strlen(buf);
    if (strrchr(buf,'\n') == buf + len - 1) {
        nlflag = 1;
        strsetsz(buf,--len);
    }

    if (strchr(buf,'\n') == buf) {
        printf("\n");
        *buf = ' ';
    }
    readcur(&row,&col);
    printf(" %s ", buf);
    readcur(&row2,&col2);
    gotoxy_(row2,col);
    revattr(col2-col);
    if (nlflag)
        printf("\n");
}


/*  error.c 
    print a reverse video error message and exit(2).
    Uses BIOS calls to get reverse video -- does not depend on ANSI driver.
    The arg list must not exceed 77 chars -- this is assumed.

    See also:
        rsay    -- rv message only.
        warning -- exit level 1
*/

#include <stdio.h>
#include <stdarg.h>
#include <core.h>
#include <string.h>

void error(char *fmt, ...)
{
    int len, row, col, row2, col2;
    char buf[80];

    va_list args;
    va_start(args, fmt);
    vsprintf(buf, fmt, args);
    va_end(args);

    len = strlen(buf);
    if (strrchr(buf,'\n') == buf + len - 1)
        strsetsz(buf,--len);

    if (strchr(buf,'\n') == buf) {
        printf("\n");
        *buf = ' ';
    }
    readcur(&row,&col);
    printf(" %s ", buf);

    readcur(&row2,&col2);
    gotoxy_(row2,col);
    revattr(col2-col);

    printf("\n");
    exit(2);
}


/*  getswi.c
    Tue Apr 17 1990

    Usage: char swiz[20];   -- big enough for all concatenated switches
           n = getswi(&argc, argv, swiz)

                    example: PROG /a/b ARG1/c/d ARG2/e/F /G
                         or  PROG -A-b ARG1/c-D ARG2-e-f -g

                    returns: swiz = "/a/b/c/d/e/f/g"
                             argc = 3 = "prog.exe", "arg1", "arg2"

                    Args and switches are always folded to lower case.
                    "/" and "-" can be mixed in any combination.

    Extracts any switches from the command line and concatenates them
    into swiz.  Also decrement argc counter so that subsequent passes
    though **argv will only extract non-switches.

    The purpose of this is to allow user to insert switches anywhere
    in the command line, but to make parsing the line simple.

    Switches are returned LOWER CASE but args are NOT modified.
*/

#include <string.h>

int getswi(int *argc, char **argv, char swiz[])
{
    int j;
    int n = 0;
    char *p;

    *swiz = '\0';
    for (j=0; j<*argc; j++)
        if ((p = strchr(argv[j],'/'))) {
            strcat(swiz, p);
            *p = '\0';
        } else if ((p = strchr(argv[j],'-'))) {
            strcat(swiz,"/");
            strcat(swiz, p+1);
            *p = '\0';
        }
    strlwr(swiz);

    n = strlen(swiz);
    if (n == 0)
        return(0);
    for (j=0; j<n; j++)
        if (swiz[j] == '-')
            swiz[j] = '/';

    /* reduce argc count for each completely removed token */
    n = 1;
    for (j=1; j<*argc; j++)
        if (strlen(argv[j]) != 0)
            argv[n++] = argv[j];
    *argc = n;

    /* count number of "/" in swiz, but leave them in there, so that */
    /* multiple switches can be discriminated */
    n = 0;
    for (j=0; j<strlen(swiz); j++)
        if (swiz[j] == '/')
            n++;
    return(n);
}


/*  flist.c
    Wed Apr 18 15:29:09 1990 John Jarvis, Austin, Texas

    Usage:  #define FILMAX 300
            char template[60];
            char *filptr[FILMAX];
            unsigned char atribz[FILMAX];

            nfil = flist(template, filptr, atribz, FILMAX);

    Stores all files except "." and ".." found in template into filptr.
    This includes attributes of normal, rdonly, hidden, system,
    volid, subdir and archive.  array atribz[] contains the attributes
    (0=normal, 16=subdir -- see dos.h).

    Template examples are "*.*", "\sou\rook\*.c"
    Sorting is not done here.  File names are returned in upper case.
*/

#include <stdio.h>
#include <core.h>
#include <string.h>
#include <dos.h>

int flist(char *template, char **filptr, unsigned char *atribz, int filmax)
{
    struct find_t fil;
    char *p;
    int len;
    int nfil = 0;

    if (_dos_findfirst(template,0xFF,&fil))
        return(0);
    if (strcmp(fil.name,".") && strcmp(fil.name,"..")) {
        len = strlen(fil.name);
        if (nfil >= filmax || (p = (char *) malloc(len+1)) == NULL)
            error("out of memory at %s at %s", template, fil.name);
        strcpy(p, fil.name);
        strupr(p);
        filptr[nfil] = p;
        atribz[nfil++] = fil.attrib;
    }

    while (_dos_findnext(&fil) == 0) {
        if (strcmp(fil.name,".") == 0 || strcmp(fil.name,"..") == 0)
            continue;
        len = strlen(fil.name);
        if (nfil >= filmax) {
            rsay("Warning: too many files in %s -- Only %d read",
               template, nfil);
            return(nfil);
        } else if ((p = (char *) malloc(len+1)) == NULL)
            error("out of memory while reading %s", template);
        strcpy(p, fil.name);
        strupr(p);
        filptr[nfil] = p;
        atribz[nfil++] = fil.attrib;
    }
    return(nfil);
}


/*  psort.c
    Tue Apr 24 1990, John Jarvis, Austin, Texas

    Usage:  int pary[LINMAX];
            char *linptr[LINMAX];
            psort(pary, linptr, nlins);

    The linptr array is sorted by qsort.  But this also returns pary
    which indicates the correct sequence of linpary if it had not been
    sorted.  It takes twice as long as normal qsort (but still goes fast).

    In practice, the linptr array will be copied from a certain element
    of a larger structure, for example a name field.  When psort is done
    sorting it, the sorted sequence in pary can be used to sort the larger
    structure in the desired order.

    The method is to do a standard qsort (manually, following K&R
    procedure, which is about 10% slower than MS qsort), but simultaneously
    keep pary updated with each swap.  This is why it takes twice as long.
    A faster method may be possible if a binary search can be done on
    the original linptr array, but this assumption may not always hold.

    This was written because I could not make qsort work on members of
    a structure. It also avoids having to swap entire structures.
    But in theory qsort should be able to handle pointers to structures
    just as it handles pointers to anything else...?

    Example:    pary[0] = 45  -- means linptr[45] is first line in order
                pary[1] = 0   -- linptr[0] is second, and so on ...
                linptr[pary[0]] == first line
*/

#include <stdio.h>
#include <stdlib.h>

void ppsort(int *pary, char **linptr, int left, int rite);
void ppswap(int *pary, char **linptr, int j, int k);

void psort(int *pary, char **linptr, int nlins)
{
    int j;

    for (j=0; j<nlins; j++)         /* default order is 0,1,2... */
        pary[j] = j;
    ppsort(pary, linptr, 0, nlins-1);
}

void ppsort(int *ip, char **vp, int left, int rite)
{
    int j, pivot;

    if (left >= rite)
        return;

    ppswap(ip, vp, left, (left + rite) / 2);
    pivot = left;           /* start with pivot on left.  The put any values */
                            /* less than pivot just to right of pivot */
    for (j = left+1; j <= rite; j++)
        if (strcmp(vp[j], vp[left]) < 0)
            ppswap(ip, vp, ++pivot, j);
    ppswap(ip, vp, left, pivot);     /* puts pivot after all lesser values */
                                     /* now recursively sort left and rite */
    ppsort(ip, vp, left, pivot-1);   /* partitions */
    ppsort(ip, vp, pivot+1, rite);
}

void ppswap(int *ip, char **vp, int j, int k)
{
    int itmp;
    char *ctmp;

    itmp = ip[j];
    ip[j] = ip[k];
    ip[k] = itmp;

    ctmp = vp[j];
    vp[j] = vp[k];
    vp[k] = ctmp;
}

/*  dirx.c
    Purpose: Extract directory name from path

    Usage: char path[40];
           char dirpath[40];
           dirx(path,dirpath);

    Example: dirx("*.*", dirpath);
                -- returns dirpath = "C:\SOU\CORE"
                                or   "C:\"   or  "D:\"

    If there is no directory in path, current working directory is
    returned.  The root directory is returned as C:\ or D:\ etc.

    Forward slash is same as backslash.

    Mon May 07, 1990, John Jarvis
    4111 Ave "A", Apt. 209, Austin, TX 78751 512-452-4755
*/

#include <stdio.h>
#include <core.h>
#include <string.h>

void dirx(char *path, char *dirpath)
{
    char buf[60];
    char *p;

    strcpy(buf,path);
    strupr(buf);
    strchg(buf,'/','\\');

    if ((p = strrchr(buf,'\\'))) {
        *p = '\0';
        if (strchr(buf,':'))
            strcpy(dirpath,buf);
        else {
            getcwd(dirpath,40);
            strcpy(dirpath+2,buf);
        }
        if (strchr(dirpath,'\\') == 0)
            strcat(dirpath,"\\");           /* root dirs */
    } else
        getcwd(dirpath,40);
}


/*  getyn.c
    Tue May 01 1990, John Jarvis, Austin, Texas

    Usage: printf("Delete this one ?");
           if (!getyn(0))
                don't delete it

    If arg is 0, prints " [N] "
    If arg is 1, prints " [Y] "

    Then gets character from keyboard, without echo.

    Prints "Yes" or "No" depending upon char and default.
    If char is "y" or "Y", return 1
    If char is "n" or "N", return 0
    Otherwise, return default.
*/

#include <conio.h>
#include <ctype.h>

int getyn(int yes1)
{
    int c;

    if (yes1)
        printf(" [Y] ");
    else
        printf(" [N] ");
    c = getch();
    if (c == 'y' || c == 'Y')
        c = 1;
    else if (c == 'n' || c == 'N')
        c = 0;
    else
        c = yes1;

    if (c)
        printf("Yes\n");
    else
        printf("No\n");

    return(c);
}
